#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<omp.h>

void scan(int N, double* a, double* b) {
  b[0] = 0;
  for (int i = 1; i < N; i++) {
    b[i] = b[i-1] + a[i-1];
  }
}

void scan_parallel(int N, double* a, double* b) {
  double * partial = calloc(omp_get_max_threads(), sizeof(double));
  #pragma omp parallel
  {
    int tid = omp_get_thread_num();
    int nthr = omp_get_num_threads();
    int chunk = N / nthr;
    int i_start = tid * chunk;
    int i_end = i_start + chunk;
    if (tid == nthr - 1) i_end = N;
    b[i_start] = 0;
    for (int i = i_start + 1; i < i_end; i++) {
      b[i] = b[i-1] + a[i-1];
    }
    for (int i = tid + 1; i < nthr; i++)
      #pragma omp atomic
      partial[i] += b[i_end - 1] + a[i_end - 1];
    #pragma omp barrier
    for (int i = i_start; i < i_end; i++) {
      b[i] += partial[tid];
    }
  }
  free(partial);
}

int main(int argc, char **argv) {
  int N = atoi(argv[1]);
  double* a = malloc(N * sizeof(double));
  double* b0 = malloc(N * sizeof(double));
  double* b1 = malloc(N * sizeof(double));
  for (int i = 0; i < N; i++)
    a[i] = rand() * 1.0 / RAND_MAX;
  scan(N, a, b0);
  scan_parallel(N, a, b1);
  double delta = 0;
  for (int i = 0; i < N; i++) {
    delta += fabs(b0[i] - b1[i]);
    printf("%f %f\n", b0[i], b1[i]);
  }
  printf("%f\n", delta);
  free(a); free(b0); free(b1);
}
